from django.db import models
from django.utils.six import add_metaclass

# Python 2/3 compatibility hackery
try:
    from itertools import ifilter as filter
except ImportError:
    pass


class AccessStateSetterMeta(type):
    """Metaclass that adds (potentially overridden) CUSTOM_STATES to ACCESS_STATES."""
    def __new__(cls, name, bases, attrs):
        """Create the new class with a complete ACCESS_STATES."""
        custom_states = attrs.get('CUSTOM_STATES', ())
        base_attr_name = 'BASE_ACCESS_STATES'
        base_states = attrs.get(base_attr_name)

        # If `attrs` does not contain BASE_ACCESS_STATES, look up the inheritance tree
        # by iterating over the elements in `bases` to find the first one that has
        # BASE_ACCESS_STATES on it.
        if not base_states:
            def has_base_states(base):
                return hasattr(base, base_attr_name)
            base_states_source = next(filter(has_base_states, bases))
            base_states = getattr(base_states_source, base_attr_name)

        attrs['ACCESS_STATES'] = custom_states + base_states
        return super(AccessStateSetterMeta, cls).__new__(cls, name, bases, attrs)


@add_metaclass(AccessStateSetterMeta)
class AccessStateExtensionMixin:
    """
    A mixin for creating extensions that add access_state to FeinCMS resources.

    This class is intended to easily build extensions that can be applied to a FeinCMS
    model to add an 'access state', which defines that model's authentication parameters.

    Pages don't have a guaranteed, defined URL, and a full site of FeinCMS pages can be
    complex to police with middleware. Adding an access state to a Page or similar FeinCMS
    object allows a middleware class to quickly and easily determine whether it should
    restrict access to that page.  MiddlewareMixin in incuna_auth.middleware.utils
    includes this functionality.

    To use this mixin, define a subclass that inherits from both this and
    feincms.extensions.Extension, and set the 'model' and 'CUSTOM_STATES' parameters.

        from feincms.module.page.models import Page
        from feincms.extensions import Extension

        class AccessState(AccessStateExtensionMixin, Extension)
            model = Page
            CUSTOM_STATES = (
                ('state1', 'The first state'),
                ('state2', 'The second state'),
            )

    You can then add it to the model (again, using Page as the example) with:

        Page.register_extensions('myapp.models.AccessState')

    The access states will include:
    - visible to all users
    - visible only to logged-in users
    - visible to everyone who can see the page's parent, or all users if no parents exist
    - any CUSTOM_STATES added by the class/application extending this mixin
    """
    model = None
    CUSTOM_STATES = ()

    STATE_ALL_ALLOWED = 'base_all'
    STATE_AUTH_ONLY = 'base_auth'
    STATE_INHERIT = 'base_inherit'
    BASE_ACCESS_STATES = (
        (STATE_ALL_ALLOWED, 'All users'),
        (STATE_AUTH_ONLY, 'Authenticated users only'),
        (STATE_INHERIT, 'Inherit from parent (allow all users if no parent exists)'),
    )

    def handle_model(self):
        """Add the ACCESS_STATES choices and the field using them to the model."""
        access_state = models.CharField(
            max_length=255,
            choices=self.ACCESS_STATES,
            default=self.STATE_INHERIT,
        )
        self.model.add_to_class('ACCESS_STATES', self.ACCESS_STATES)
        self.model.add_to_class('access_state', access_state)

    def handle_modeladmin(self, modeladmin):
        """Ensure the model admin gets the access state option too."""
        modeladmin.add_extension_options('access_state')



import os

import dj_database_url


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DEBUG = True
ALLOWED_HOSTS = []
ROOT_URLCONF = 'incuna_auth.urls'
STATIC_URL = '/static/'

SECRET_KEY = 'not-for-production'
PASSWORD_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher',)

DATABASES = {
    'default': dj_database_url.config(default='postgres://localhost/incuna_auth')
}
DEFAULT_FILE_STORAGE = 'inmemorystorage.InMemoryStorage'

INSTALLED_APPS = (
    'incuna_auth',
    'incuna_auth.tests',

    'incuna_test_utils',

    # Work around 'relation does not exist' errors by ordering the installed apps:
    #   contenttypes -> auth -> everything else.
    # See: https://code.djangoproject.com/ticket/10827#comment:12
    #      http://stackoverflow.com/q/29689365
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.sessions',
)
MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)
REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': ('rest_framework.permissions.IsAuthenticated',),
    'DEFAULT_RENDERER_CLASSES': ('rest_framework.renderers.JSONRenderer',),
}

# LoginRequiredMiddleware data
LOGIN_REQUIRED_SEND_MESSAGE = False

# BasicAuthenticationMiddleware data
BASIC_WWW_AUTHENTICATION_USERNAME = 'user'
BASIC_WWW_AUTHENTICATION_PASSWORD = 'pass'
BASIC_WWW_AUTHENTICATION = True

# Additional translation folders.
LOCALE_PATHS = (
    os.path.join(BASE_DIR, 'incuna_auth', 'locale'),
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.core.context_processors.request',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

TEST_RUNNER = 'test_project.test_runner.Runner'

TRANSLATE_URLS = bool(os.environ.get('TRANSLATE_URLS', False))



from crispy_forms.helper import FormHelper
from crispy_forms.layout import (
    Div,
    Field,
    Layout,
    Submit,
)
from django.contrib.auth.forms import PasswordResetForm


class CrispyPasswordResetForm(PasswordResetForm):
    helper = FormHelper()
    helper.layout = Layout(
        Div(
            Field('email'),
            Submit('submit', 'Reset'),
        )
    )



from unittest import skipIf, skipUnless, TestCase

import django
from django.conf import settings
from django.contrib.auth import views
from django.core.urlresolvers import resolve, reverse
from django.utils import translation
from django.views.generic import RedirectView


class URLsMixin(object):
    """
    A TestCase Mixin with a check_url helper method for testing urls.
    Pirated with slight modifications from incuna_test_utils
    https://github.com/incuna/incuna-test-utils/blob/master/incuna_test_utils/testcases/urls.py
    """

    def check_url(self, view_method, expected_url, url_name,
                  url_args=None, url_kwargs=None):
        """
        Assert a view's url is correctly configured

        Check the url_name reverses to give a correctly formated expected_url.
        Check the expected_url resolves to the correct view.
        """

        reversed_url = reverse(url_name, args=url_args, kwargs=url_kwargs)
        self.assertEqual(reversed_url, expected_url)

        # Look for a method rather than a class here
        # (just because of what we're testing)
        resolved_view_method = resolve(expected_url).func
        self.assertEqual(resolved_view_method.__name__, view_method.__name__)


class TestURLs(URLsMixin, TestCase):

    def test_login(self):
        self.check_url(
            views.login,
            '/login/',
            'login',
        )

    @skipUnless(settings.TRANSLATE_URLS, 'Only run if TRANSLATE_URLS=True')
    def test_login_translate_enabled(self):
        with translation.override('de_AT'):
            self.check_url(
                views.login,
                '/Anmeldung/',
                'login',
            )

    @skipIf(settings.TRANSLATE_URLS, 'Only run if TRANSLATE_URLS=False')
    def test_login_translate_disabeled(self):
        with translation.override('de_AT'):
            self.check_url(
                views.login,
                '/login/',
                'login',
            )

    def test_logout(self):
        self.check_url(
            views.logout,
            '/logout/',
            'logout',
        )

    def test_password_change(self):
        self.check_url(
            views.password_change,
            '/password/change/',
            'password_change',
        )

    def test_password_change_done(self):
        self.check_url(
            views.password_change_done,
            '/password/change/done/',
            'password_change_done',
        )

    def test_password_reset(self):
        self.check_url(
            views.password_reset,
            '/password/reset/',
            'password_reset',
        )

    def test_password_reset_done(self):
        self.check_url(
            views.password_reset_done,
            '/password/reset/done/',
            'password_reset_done',
        )

    def test_password_reset_complete(self):
        self.check_url(
            views.password_reset_complete,
            '/password/reset/complete/',
            'password_reset_complete',
        )

    def test_sso(self):
        self.check_url(
            RedirectView.as_view(),
            '/sso/',
            'sso_login',
        )

    @skipIf(django.VERSION < (1, 6), 'This test is for a different Django version')
    def test_password_reset_confirm_uidb64(self):
        uidb64 = '09_AZ-az'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm,
            '/password/reset/confirm/{0}/{1}/'.format(uidb64, token),
            'password_reset_confirm',
            url_kwargs={'uidb64': uidb64,
                        'token': token, },
        )

    @skipUnless(
        (1, 6) >= django.VERSION > (1, 7),
        'This test is for a different Django version',
    )
    def test_password_reset_confirm_uidb36(self):
        uidb36 = '09AZaz'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm_uidb36,
            '/password/reset/confirm/{0}-{1}/'.format(uidb36, token),
            'password_reset_confirm',
            url_kwargs={'uidb36': uidb36,
                        'token': token, },
        )

    @skipIf(django.VERSION >= (1, 6), 'This test is for a different Django version')
    def test_password_reset_confirm_dj1_5(self):
        uidb36 = '09AZaz'
        token = '09AZaz-09AZaz'
        self.check_url(
            views.password_reset_confirm,
            '/password/reset/confirm/{0}-{1}/'.format(uidb36, token),
            'password_reset_confirm',
            url_kwargs={'uidb36': uidb36,
                        'token': token, },
        )



import re

# Python 2/3 compatibility hackery
try:
    unicode
except NameError:
    unicode = str


def compile_url(url):
    clean_url = unicode(url).lstrip(u'/')
    return re.compile(clean_url)


def compile_urls(urls):
    return [compile_url(expr) for expr in urls]



import mock

from incuna_auth.middleware import permission_feincms
from incuna_auth.models import AccessStateExtensionMixin as AccessState
from .utils import RequestTestCase


class TestFeinCMSPermissionMiddleware(RequestTestCase):
    middleware_class = permission_feincms.FeinCMSPermissionMiddleware
    middleware_path = (
        'incuna_auth.middleware.permission_feincms.' +
        'FeinCMSPermissionMiddleware.{}'
    )
    get_page_method = middleware_path.format('_get_page_from_path')

    CUSTOM_STATE = ('custom', 'Custom state')

    class DummyFeinCMSPage:
        def __init__(self, access_state):
            self.access_state = access_state

    def setUp(self):
        self.middleware = self.middleware_class()

    def make_request(self, access_state=CUSTOM_STATE, **kwargs):
        """Create a request with a suitable feincms_page attribute."""
        request = self.create_request(**kwargs)
        request.feincms_page = self.DummyFeinCMSPage(access_state=access_state)
        return request

    def test_get_protected_states(self):
        """Assert the default return value of this method."""
        expected_states = [AccessState.STATE_AUTH_ONLY]
        self.assertEqual(expected_states, self.middleware.get_protected_states())

    def test_get_resource_access_state(self):
        """Assert that the correct access_state value comes back from the request."""
        request = self.make_request()
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            access_state = self.middleware._get_resource_access_state(request)

        expected_state = self.CUSTOM_STATE
        self.assertEqual(expected_state, access_state)

    def test_get_resource_access_state_no_page(self):
        """Assert that the access_state is None when feincms_page is None."""
        request = self.make_request()
        with mock.patch(self.get_page_method, return_value=None):
            access_state = self.middleware._get_resource_access_state(request)

        self.assertIsNone(access_state)

    def test_get_resource_access_state_unrestricted(self):
        """
        Assert that the certain access_state values are never restricted.

        STATE_INHERIT is only unrestricted when the resource has no parent; the
        behaviour of STATE_INHERIT on a resource *with* a parent is tested elsewhere, so
        both cases are covered.
        """
        unrestricted_states = (AccessState.STATE_INHERIT, AccessState.STATE_ALL_ALLOWED)
        request = self.make_request()
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            for state in unrestricted_states:
                request.feincms_page.access_state = state
                request.feincms_page.parent = None
                self.assertIsNone(self.middleware._get_resource_access_state(request))

    def test_get_resource_access_state_inherited(self):
        """Assert that the method deals correctly with inherited access_state values."""
        parent_page = self.DummyFeinCMSPage(self.CUSTOM_STATE)
        request = self.make_request(access_state=AccessState.STATE_INHERIT)
        request.feincms_page.parent = parent_page

        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            access_state = self.middleware._get_resource_access_state(request)

        expected_state = self.CUSTOM_STATE
        self.assertEqual(expected_state, access_state)

    def test_get_resource_access_state_null_parent(self):
        """
        Assert that the method doesn't explode when the Page has a parent which is None.
        """
        request = self.make_request(access_state=AccessState.STATE_INHERIT)
        request.feincms_page.parent = None

        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            access_state = self.middleware._get_resource_access_state(request)

        self.assertIsNone(access_state)

    def test_resource_protected(self):
        """
        Assert that the URL is protected when the value of get_protected_states contains
        feincms_page.access_state.
        """
        request = self.make_request()
        get_states_method = self.middleware_path.format('get_protected_states')
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            with mock.patch(get_states_method, return_value=[self.CUSTOM_STATE]):
                self.assertTrue(self.middleware.is_resource_protected(request))

    def test_resource_unprotected(self):
        """
        Assert that the URL is not protected when the value of get_protected_states
        doesn't contain feincms_page.access_state.
        """
        request = self.make_request()
        get_states_method = self.middleware_path.format('get_protected_states')
        state = ('other-state', 'Other State')
        with mock.patch(self.get_page_method, return_value=request.feincms_page):
            with mock.patch(get_states_method, return_value=[state]):
                self.assertFalse(self.middleware.is_resource_protected(request))



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .permission import LoginPermissionMiddlewareMixin, UrlPermissionMiddleware
from .utils import compile_urls


def check_request_has_user():
    """
    Check that LoginRequiredMiddleware isn't being used without its dependency.

    LoginRequiredMiddleware needs django.contrib.auth.middleware.AuthenticationMiddleware.
    """
    middlewares = settings.MIDDLEWARE_CLASSES

    if 'django.contrib.auth.middleware.AuthenticationMiddleware' in middlewares:
        return

    error_message = ' '.join((
        "MIDDLEWARE_CLASSES does not contain AuthenticationMiddleware.",
        "LoginRequiredMiddleware requires authentication middleware to be",
        "installed. Ensure that your MIDDLEWARE_CLASSES setting includes",
        "'django.contrib.auth.middleware.AuthenticationMiddleware'.",
    ))
    raise ImproperlyConfigured(error_message)


class LoginRequiredMiddleware(LoginPermissionMiddlewareMixin, UrlPermissionMiddleware):
    """
    Middleware that requires a user to be authenticated.

    Middleware that requires a user to be authenticated to view any page in
    LOGIN_PROTECTED_URLS other than LOGIN_URL. Exemptions to this requirement
    can optionally be specified in settings via a list of regular expressions
    in LOGIN_EXEMPT_URLS (which you can copy from your urls.py).

    Requires authentication middleware and template context processors to be
    loaded. You'll get an error if they aren't.

    Will default to protecting everything if LOGIN_PROTECTED_URLS is not in
    settings.

    Original code from:
    http://onecreativeblog.com/post/59051248/django-login-required-middleware

    This version has been modified to allow us to define areas of the site to
    password protect instead of protecting everything under /.
    """
    base_unauthorised_redirect_url = settings.LOGIN_URL

    login_exempt_urls = [settings.LOGIN_URL, settings.LOGOUT_URL]
    login_exempt_urls += getattr(settings, 'LOGIN_EXEMPT_URLS', [])
    login_protected_urls = getattr(settings, 'LOGIN_PROTECTED_URLS', [r'^'])

    EXEMPT_URLS = compile_urls(login_exempt_urls)
    PROTECTED_URLS = compile_urls(login_protected_urls)

    def __init__(self, check=True):
        if check:
            check_request_has_user()

    def get_exempt_url_patterns(self):
        return self.EXEMPT_URLS

    def get_protected_url_patterns(self):
        return self.PROTECTED_URLS



from unittest import TestCase

from crispy_forms.layout import Field, Submit

from incuna_auth import forms


class TestCrispyPasswordResetForm(TestCase):

    def test_form(self):
        form = forms.CrispyPasswordResetForm()
        layout = form.helper.layout

        # The shape of the tree of elements overall
        self.assertEqual(len(layout), 1)
        div = layout[0]
        self.assertEqual(len(div), 2)

        # The Field element
        self.assertTrue(isinstance(div[0], Field))
        self.assertEqual(div[0].fields, ['email'])

        # The Submit element
        self.assertTrue(isinstance(div[1], Submit))
        self.assertEqual(div[1].name, 'submit')
        self.assertEqual(div[1].value, 'Reset')



from django.conf import settings

from .permission import LoginPermissionMiddlewareMixin
from .permission_feincms import FeinCMSPermissionMiddleware


class FeinCMSLoginRequiredMiddleware(
    LoginPermissionMiddlewareMixin,
    FeinCMSPermissionMiddleware
):
    """
    Middleware that requires a user to be authenticated to view any page with an
    access_state of STATE_AUTH_ONLY.

    Requires authentication middleware, template context processors, and FeinCMS's
    add_page_if_missing middleware to be loaded. You'll get an error if they aren't.
    """
    base_unauthorised_redirect_url = settings.LOGIN_URL



import django
from django.conf import settings
from django.conf.urls import url
from django.contrib.auth.views import (
    login,
    logout,
    password_change,
    password_change_done,
    password_reset,
    password_reset_complete,
    password_reset_confirm,
    password_reset_done,
)
from django.core.urlresolvers import get_callable, reverse_lazy
from django.utils.translation import ugettext_lazy
from django.views.generic import RedirectView


# Only translate the urls if `TRANSLATE_URLS` is `True`.
if getattr(settings, 'TRANSLATE_URLS', False):
    _ = ugettext_lazy
else:
    def _(s):
        return s


auth_login_form = getattr(
    settings,
    'INCUNA_AUTH_LOGIN_FORM',
    'django.contrib.auth.forms.AuthenticationForm',
)
auth_form = get_callable(auth_login_form)

password_reset_form = getattr(
    settings,
    'INCUNA_PASSWORD_RESET_FORM',
    'incuna_auth.forms.CrispyPasswordResetForm',
)
reset_form = get_callable(password_reset_form)


urlpatterns = [
    url(
        _(r'^login/$'),
        login,
        {'authentication_form': auth_form},
        name='login',
    ),
    url(
        _(r'^logout/$'),
        logout,
        {'template_name': 'registration/logout.html'},
        name='logout',
    ),
    # Change password (when logged in)
    url(
        _(r'^password/change/$'),
        password_change,
        name='password_change',
    ),
    url(
        _(r'^password/change/done/$'),
        password_change_done,
        name='password_change_done',
    ),
    # Reset password (when not logged in via unique email link)
    url(
        _(r'^password/reset/$'),
        password_reset,
        {'password_reset_form': reset_form},
        name='password_reset',
    ),
    url(
        _(r'^password/reset/done/$'),
        password_reset_done,
        name='password_reset_done',
    ),
    url(
        _(r'^password/reset/complete/$'),
        password_reset_complete,
        name='password_reset_complete',
    ),
    url(
        _(r'^sso/$'),
        RedirectView.as_view(url=reverse_lazy('admin:admin_sso_openiduser_start')),
        name='sso_login',
    ),
]

if django.VERSION >= (1, 6):
    urlpatterns += [
        url(
            _(
                r'^password/reset/confirm/' +
                r'(?P<uidb64>[0-9A-Za-z_\-]+)/' +
                r'(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'
            ),
            password_reset_confirm,
            name='password_reset_confirm',
        ),
    ]

if django.VERSION < (1, 7):
    urlpatterns += [
        url(
            _(
                r'^password/reset/confirm/' +
                r'(?P<uidb36>[0-9A-Za-z]{1,13})' +
                r'-(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,20})/$'
            ),
            password_reset_confirm,
            name='password_reset_confirm',
        ),
    ]
